# fancywin.tcl: some general fancy toplevel window config.
# Copyright (C) 2010 Jeroen Hoppenbrouwers

namespace eval fcyw {
}; # fcyw


proc fcyw::init {handle args} {
  # Add fancy features to a toplevel window's system menu (top-left icon).
  # <in> handle = A TK toplevel window name.
  if {![winfo exists $handle]} return

  array set options {
    -transparentcolor {}
  }
  foreach {option value} $args {
    set options($option) $value
  }

  # Does this window already have a menu bar?
  set menubar [$handle cget -menu]
  if {$menubar==""} {
    # Not yet. Create one and add the system menu.
    set menubar [menu $handle.menubar]
    $handle config -menu $menubar
    set m [menu $menubar.system -tearoff 0]
    $menubar add cascade -menu $m
  } else {
    # Already exists. Just pick up the system menu (magic name).
    set m $menubar.system
  }

  # Add some items to the window's System menu.
  $m add separator
  $m add command -label "Window Config" -underline 0 \
      -command "fcyw::config $handle"

  # Fetch the settings for this window.
  array set fcywPrefs [prefs::get fcyw.cfg]

  # Make sure there are sane defaults ...
  if {[info exists fcywPrefs($handle.alpha)]} {
    set alpha $fcywPrefs($handle.alpha)
  } else {
    set alpha 1.0
    set fcywPrefs($handle.alpha) $alpha
  }
  if {[info exists fcywPrefs($handle.topmost)]} {
    set topmost $fcywPrefs($handle.topmost)
  } else {
    set topmost 0
    set fcywPrefs($handle.topmost) $topmost
  }
  if {[info exists fcywPrefs($handle.toolwindow)]} {
    set toolwindow $fcywPrefs($handle.toolwindow)
  } else {
    set toolwindow 0
    set fcywPrefs($handle.toolwindow) $toolwindow
  }
  if {[info exists fcywPrefs($handle.nobackground)]} {
    set nobackground $fcywPrefs($handle.nobackground)
  } else {
    set nobackground 0
    set fcywPrefs($handle.nobackground) $nobackground
  }
  # ... and write these immediately back to disk for safety.
  prefs::put fcyw.cfg [array get fcywPrefs]

  wm attributes $handle -alpha $alpha -topmost $topmost -toolwindow $toolwindow
  if {($options(-transparentcolor)!="") && $nobackground} {
    wm attributes $handle -transparentcolor $options(-transparentcolor)
  }
}; # fancyWindow


proc fcyw::config {handle} {
  # Pop open the fancy configuration dialog for the toplevel window.
  # <in> handle = A TK toplevel window name.
  if {![winfo exists $handle]} return

  # Create yet another toplevel window.
  global APPNAME
  variable _CONFIG

  ##### Setup for a modal simple dialog.
  set w [toplevel $handle.fcywcfg -class Dialog]
  wm title $w "$APPNAME Window Config"
  wm transient $w .
  wm protocol $w WM_DELETE_WINDOW "$w.b.cancel invoke"
  wm resizable $w 0 0
  # Some attributes are inherited from the parent, which isn't good here.
  wm attributes $w -alpha 1.0 -topmost 1
  centerParent $w

  ##### Get current values.
  set previousAlpha [wm attributes $w -alpha]
  array set fcywPrefs [prefs::get fcyw.cfg]
  set _CONFIG(transp)       [expr 100-($fcywPrefs($handle.alpha)*100)]
  set _CONFIG(topmost)      $fcywPrefs($handle.topmost)
  set _CONFIG(toolwindow)   $fcywPrefs($handle.toolwindow)
  set _CONFIG(nobackground) $fcywPrefs($handle.nobackground)

  ##### Config items.
  pack [frame $w.t] -fill x -side top -pady 10

  pack [frame $w.t.1] -expand 1 -fill both -side top -pady 0 -padx 5
  pack [label $w.t.1.l -text "Transparent %" -width 12 -anchor w] -side left
  pack [scale $w.t.1.s -orient horizontal -length 100 -to 80 \
    -variable fcyw::_CONFIG(transp) \
    -command "fcyw::setAlpha $handle"] -side left -padx 2 -fill x

  pack [frame $w.t.2] -expand 1 -fill both -side top -pady 0 -padx 5
  pack [label $w.t.2.l -text "Stay On Top" -width 12 -anchor w] -side left
  pack [checkbutton $w.t.2.c -variable fcyw::_CONFIG(topmost)] \
    -side left -padx 2 -fill x

  pack [frame $w.t.3] -expand 1 -fill both -side top -pady 0 -padx 5
  pack [label $w.t.3.l -text "Tool Window" -width 12 -anchor w] -side left
  pack [checkbutton $w.t.3.c -variable fcyw::_CONFIG(toolwindow) \
            -text "(close window after change)"] \
    -side left -padx 2 -fill x

  pack [frame $w.t.4] -expand 1 -fill both -side top -pady 0 -padx 5
  pack [label $w.t.4.l -text "No Background" -width 12 -anchor w] -side left
  pack [checkbutton $w.t.4.c -variable fcyw::_CONFIG(nobackground) \
            -text "(close window after change)"] \
    -side left -padx 2 -fill x

  ##### Buttons.
  frame $w.b
  pack $w.b -fill x -side top
  button $w.b.ok -text "Ok" -underline 0 -width 10 \
    -command "set fcyw::_CONFIG(button) ok"
  button $w.b.cancel -text "Cancel" -underline 0 -width 10 \
    -command "set fcyw::_CONFIG(button) cancel"
  pack $w.b.cancel $w.b.ok -side right -padx 2 -pady 2
  bind $w <Return> "$w.b.ok invoke"
  bind $w <Escape> "$w.b.cancel invoke"

  raise $w
  focus $w
  grab $w
  update

  # Wait for the user to click a button.
  vwait fcyw::_CONFIG(button)
  if {$fcyw::_CONFIG(button)=="ok"} {
    # Make some selections immediately active in the application.
    wm attributes $handle -topmost    $fcyw::_CONFIG(topmost)
    # Update the complete fancywin preferences array.  
    set fcywPrefs($handle.alpha) [expr (100-$fcyw::_CONFIG(transp))/100.0]
    set fcywPrefs($handle.topmost)      $fcyw::_CONFIG(topmost)
    set fcywPrefs($handle.toolwindow)   $fcyw::_CONFIG(toolwindow)
    set fcywPrefs($handle.nobackground) $fcyw::_CONFIG(nobackground)
    prefs::put fcyw.cfg [array get fcywPrefs]
  } else {
    # Cancel.
    # Reset alpha to what it was before the change (if any).
    wm attributes $handle -alpha $previousAlpha
  }

  destroy $w
}; # config


proc fcyw::setAlpha {handle alphaPercent} {
  wm attributes $handle -alpha [expr (100-$alphaPercent)/100.0]
}; # setAlpha

# end fancywin.tcl
