# generic.tcl: Generic Procedures for Tcl scripts.
# Copyright (C) 2008 Jeroen Hoppenbrouwers

set RANDOM 7654321
array set SLEEPTIMER {}


proc die {msg} {
  # End the program with a message.
  # <in> msg = The message to put out (as a background error).
  global errorInfo
  set errorInfo "ERROR: $msg"
  bgerror "$msg\n\nThe program will be terminated."
  exit 1
}; # die


proc rnd {max} {
  # Calculates a semi-random integer between and including 0 and max.
  # <in> max = The maximum outcome of the calculation, inclusive. May be
  #            a negative value, result will then be negative or zero.
  # <out> Random integer.
  global RANDOM
  set RANDOM [expr ([clock clicks]+$RANDOM*[clock clicks])%9999999]
  return [expr $RANDOM % ($max+1)]
}; # rnd


proc rndshift {center shift} {
  # Produces a random number in the range (area +/- shift), inclusive.
  # No real Gaussian curve, yet.
  # <in> center = The "top" of the output distribution.
  # <in> shift = The maximum deviation in both numerical directions.
  # <out> Random integer.
  set shift [rnd $shift]
  if {[rnd 1]} {
    return [expr $center+$shift]
  } else {
    return [expr $center-$shift]
  }
}; # rndshift


proc rndpadd {value maxperc} {
  # Produces a random number in the range (input value + increase), where
  # increase is random between zero and the given percentage of value.
  # I.e, {50 20} produces something between 50 and 60, inclusive.
  # <in> value = The base value to start from.
  # <in> maxperc = The percentage that determines the maximum random addition.
  #                This value may be given as a negative one, which leads to
  #                an output between 40 and 50 in the above example.
  # <out> Random integer.
  set perc [rnd $maxperc]
  set shift [expr $value*($perc.0/100)]
  return [expr round($value+$shift)]
}; # rndpadd


proc rndlist {list} {
  # Returns one of the elements out of the given list at random.
  # <in> list = Any list.
  # <out> A random element out of the list, or {} if the list is empty.
  if {$list=={}} {return {}}
  return [lindex $list [rnd [expr [llength $list]-1]]]
}; # rndlist


proc min {a b} {
  # Return the lowest value of the two given.
  if {$a<$b} {
    return $a
  } else {
    return $b
  }
}; # min


proc max {a b} {
  # Return the highest value of the two given.
  if {$a>$b} {
    return $a
  } else {
    return $b
  }
}; # max


proc d {val} {
  # <in> Any value, possibly with leading zeroes (octal problems!).
  # <out> The same value, cleaned up so that it is guaranteed in
  # proper decimal format.
  set s ""
  set i [string trim $val]
  if {[string index $i 0]=="-"} {
    set s "-"
    set i [string range $i 1 end]
  }
  set i [string trimleft $i "0"]
  if {$i==""} {return 0}
  return [expr "$s$i"]
}; # d


proc compass {deg} {
  # Return a three-digit compass course, always a multiple of ten.
  # <in> deg = Any number between -360 and 719.
  # <out> The same course rounded to a multiple of ten degrees, always in
  #       three digits, and between 000 and 350.
  set deg [expr round($deg/10)*10]
  if {$deg<0} {
    incr deg 360
  }
  if {$deg>359} {
    incr deg -360
  }
  return [format "%03i" $deg]
}; # compass


proc wrap {msg {column 78} {linefeeds 0}} {
  # Wraps the given string at n columns wide. Wraps only between words.
  # <in> msg = The string to wrap.
  # <in> column = Optionally, the column to wrap (max line length).
  #               Default is 78.
  # <in> linefeeds = Optionally, existing line feeds are retained.
  #                  If 0, all line feeds are deleted first.
  # <out> A string including line feeds before or at the n-th column.

  # Delete all line feeds (to stuff as much text on the screen as possible)?
  if {!$linefeeds} {regsub -all "\n" $msg " " msg}

  # Run over all existing lines.
  set out {}
  foreach line [split $msg "\n"] {
    while {[string length $line]>$column} {
      # Line too long, break it.
      set break [string wordstart $line $column]
      if {$break==0} {
	# No spaces in the whole line? Ah well, then break at column.
	set break $column
      }
      append out "[string trim [string range $line 0 [expr $break-1]]]\n"
      set line [string trim [string range $line $break end]]
    }
    append out "[string trim $line]\n"
  }
  return $out
}; # wrap


proc dump {varName {ch stderr}} {
  # Dumps a variable on stderr. The variable might be
  # a scalar or an array; the latter one is dumped in nice table layout.
  # <in> varName = The variable to dump.
  # <in> ch = Optionally, the channel to dump the variable to. Default
  #           is stderr.
  upvar $varName var

  # Check for variable existance.
  if {![info exists var]} {
    puts $ch "DUMP: unknown variable $varName\n"
    return
  }

  # Find out if the variable is scalar or array.
  if {[array size var]==0} {
    # Scalar (or empty array).
    puts $ch "DUMP: scalar $varName = `$var'"
  } else {
    # Array.
    puts $ch "DUMP: array $varName ="
    foreach i [lsort [array names var]] {
      puts $ch "  ($i) = `$var($i)'"
    }
  }; # if scalar or array
}; # dump


proc centerScreen {w} {
  # Center the window on the screen.
  # <in> w = A toplevel window to be centered on the screen.
  wm withdraw $w
  update idletasks
  set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2]
  set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2]
  wm geom $w +$x+$y
  update idletasks
  wm deiconify $w
}; # centerScreen


proc centerParent {w} {
  # Center the window on its parent.
  # <in> w = A toplevel window to be centered on its parent.
  set p [winfo parent $w]
  if {$p==""} {
    # No parent -- ignore the request.
    return
  }
  wm withdraw $w
  update idletasks
  set rootx [winfo rootx $p]
  set rooty [winfo rooty $p]
  set x [expr [winfo width $p]/2 - [winfo reqwidth $w]/2]
  set y [expr ([winfo height $p]-20)/2 - [winfo reqheight $w]/2]
  wm geom $w +[expr $rootx+$x]+[expr $rooty+$y]
  update idletasks
  wm deiconify $w
}; # centerParent


proc bgerror {msg} {
  # Gives a stack trace on stdout when an event handler produces
  # an error.
  global errorInfo ETCDIR VERSION APPNAME

  # Dump the error in a log file for later research.
  set f [open "$ETCDIR/debug.log" a]
  puts $f "BACKGROUND ERROR"
  puts $f "[clock format [clock seconds]] V$VERSION"
  puts $f $errorInfo
  puts $f "END BACKGROUND ERROR\n"
  close $f

  # Dump it also on stdout, which won't be visible on Windows.
  puts "\nBACKGROUND ERROR"
  puts $errorInfo
  puts "END BACKGROUND ERROR\n"

  # Do a popup to inform the user.
  tk_messageBox -icon error -message "An error occurred. Please report it\
      to hoppie@hoppie.nl\
      \nLook in the etc/debug.log file for a more detailed description.\n\
      \n$msg" -title "Error in $APPNAME"\
      -type ok
  return
}; # bgerror


proc hhmm2min {hhmm} {
  # Convert a string HHMM to minutes since midnight.
  set hhmm [format %04d [d $hhmm]]
  set hrs [d [string range $hhmm 0 1]]
  set min [d [string range $hhmm 2 3]]
  return [expr ($hrs*60)+$min]
}

proc min2hhmm {min} {
  # Convert an integer with minutes since midnight to a string HHMM.
  set min [expr $min%(24*60)]
  set hrs [expr $min/60]
  set min [expr $min%60]
  return [format %02d%02d $hrs $min]
}


proc dset {var value default} {
  # Sets a variable to value, unless value is "", then use default.
  upvar $var myVar
  if {$value==""} {
    set myVar $default
  } else {
    set myVar $value
  }
}


proc sleep {seconds} {
  # Pauses the execution for the given amount of seconds while keeping
  # the event loop alive, as opposite to the "after" command. Is fully
  # reentrant, can handle multiple sleeps nested and tangled.
  global SLEEPTIMER

  # Get the next timer ID.
  set id [expr [lindex [lsort [array names SLEEPTIMER]] end]+1]
  if {$id>32000} {set id 0}
  set SLEEPTIMER($id) 1

  after [expr round($seconds*1000)] "set SLEEPTIMER($id) 0"
  vwait SLEEPTIMER($id)

  # Clean up.
  unset SLEEPTIMER($id)
}; # sleep

# end generic.tcl
