# prefs.tcl: persistent user preferences.

# Copyright (C) 2008 Jeroen Hoppenbrouwers <hoppie@hoppie.nl>
# This program is free software; more information in the LICENSE file.


namespace eval prefs {
  variable PREFSPATH "$ETCDIR/settings.ini"

  variable LAZYWRITE 1000       ; # Lazy write time in milliseconds.
  variable PREFS                ; # Actual preferences.
  variable DEFAULTS             ; # Preference defaults and allowed keys.
  array set DEFAULTS {}
}


proc prefs::init {{appName "settings"} {defaults {}}} {
  # Initializes the data structures. May be used to set application name and
  # preference defaults. Both are recommended.
  # Note that the storage of the keys on disk has no information about
  # defaults or the mandatory presence of a key.
  # <in> appName = Optionally, the application name, which will be used to
  #      determine the preferences file name. Defaults to "settings",
  #      leading to "etc/settings.ini".
  # <in> defaults = Optionally, list {key value key value...} of all possible
  #      settings and their defaults. If left empty, no defaults are present
  #      (unknown keys return "") and any key name is accepted. If present,
  #      any unknown key leads to a fatal error.
  global   ETCDIR
  variable PREFSPATH
  variable PREFS
  variable DEFAULTS

  set PREFSPATH "$ETCDIR/$appName.ini"
  array set DEFAULTS $defaults
}; # init


proc prefs::done {} {
  # Flush the preferences to disk, nothing else yet.
  flush
}; # done


proc prefs::put {key value} {
  # Stores a preference setting. In case default preferences are known, a
  # check is done on the existance of the name. A fatal error occurs if an
  # unknown setting key is used.
  # <in> key = The key to set.
  # <in> value = The value to set.
  variable PREFS
  variable LAZYWRITE
  variable DEFAULTS

  # Cancel a possible pending flush trigger.
  after cancel {prefs::flush}

  # Check whether key name enforcement is required.
  if {[llength [array names DEFAULTS]]>0} {
    # It is.
    if {![info exists DEFAULTS($key)]} {
      # Unknown key, bail out.
      die "Unknown preference '$key'."
    }
  }

  # Policy: default values are only used during retrieval. This is to
  # prevent a program suddenly changing defaults with a new version while a
  # user has previous settings on disk already. If program logic treats ""
  # as default, a change may still occur.
  if {$value==""} {
    if {[info exists PREFS($key)]} {unset PREFS($key)}
  } else {
    set PREFS($key) $value
  }

  # Start a flush timer. This causes subsequent settings to be flushed only
  # after a specific lazy time.
  after $LAZYWRITE {prefs::flush}
}; # put


proc prefs::get {key} {
  # Retrieve a preference setting.
  # <in> key = The setting to retrieve.
  # <out> The value of the setting, or "" if the setting was not found.
  #       If the setting has a default, the default is returned instead of
  #       "" (obviously the default may be "" again).
  variable PREFSPATH
  variable PREFS
  variable DEFAULTS

  # Check whether key name enforcement is required.
  if {[llength [array names DEFAULTS]]>0} {
    # It is.
    if {![info exists DEFAULTS($key)]} {
      # Unknown key, bail out.
      die "Unknown preference '$key'."
    }
  }

  # Check for existence of the PREFS array.
  if {![info exists PREFS]} {
    # Array does not exist, try to load the preferences file. Skip on error.
    if {[catch {open $PREFSPATH} f]} {
      # Most likely the file does not yet exist. Ignore.
      # We cannot notify the user of this event (no log target).
    } else {
      array set PREFS [read $f]
      close $f
    }

    # Loading the prefs db succeeded.
    # It is possible to load up an older .INI file which contains keys that
    # have in the mean time been discontinued. This can only be detected if
    # an array of defaults is present (and populated).
    if {[llength [array names DEFAULTS]]>0} {
      foreach newkey [array names PREFS] {
        if {![info exists DEFAULTS($newkey)]} {
          # This key has been discontinued. Remove it from memory.
          unset PREFS($newkey)
        }
      }
    }
  }; # if prefsdb needs loading

  # Check for existence of the required key.
  if {![info exists PREFS($key)]} {
    # Does not exist in the array. Does it have a default?
    if {[info exists DEFAULTS($key)]} {
      # It has a default; return the default.
      return $DEFAULTS($key)
    } else {
      # No default, return the empty string.
      return ""
    }
  }

  # Existing key name with no default required. Return the key's value.
  return $PREFS($key)
}; # get


proc prefs::flush {} {
  # Write the settings buffer to disk. The defaults buffer is never written
  # to disk, it remains fully under control of the program.
  variable PREFSPATH
  variable PREFS

  if {[info exists PREFS]} {
    if {![catch {open $PREFSPATH w} f]} {
      foreach p [lsort [array names PREFS]] {
        puts $f [list $p $PREFS($p)]
      }
      close $f
    }; # catch
  }; # exists
}; # flush


proc prefs::exist {} {
  # Check wether the preferences file exists.
  # <out> 0 when the prefs do not exist, 1 if they do exist.
  variable PREFSPATH
  return [file exists $PREFSPATH]
}; # exist

# [EOF]
