# External MCDU library. Depends on the Broker connector if so configured
# with xmcdu::init. Provides generic MCDU functions, but no true page
# handlers of any kind.

namespace eval xmcdu {
  variable MCDU "";   # Holds stream ID for the MCDU connection, if any.
  variable MODULES;   # Array; pending registration attempts.
                      # (modName) = {callback pending}.
  variable PASSTHROUGH; # Array; passthrough requests per module.
                      # (modName) = {1 when passthrough, 0 when not}.
  variable SERVER ""; # When "", uses the Broker library. When not "", this
                      # variable contains "hostname:port" for direct conn.
}; # xmcdu


##### MCDU Communication ################################################

proc xmcdu::init {modName {event ""} {passthrough ""}} {
  # Set up a connection to the external MCDU. If no acknowledgement
  # comes in 30 seconds, retry until the xmcdu::done is given.
  # This procedure needs to keep a list of currently registered modules,
  # in case of an MCDU reset where they all need to be re-registered.
  # <in> modName = The module name to be registered at the MCDU.
  # <in> event = The fully-qualified procedure name to handle events
  #              for the module. Required for a first registration attempt,
  #              ignored for subsequent registration checks.
  # <in> passthrough = When given, puts the module in passthrough mode so it
  #              will receive raw keystrokes.
  variable MODULES
  variable PASSTHROUGH

  # Kill possible pending init attempts for this module.
  after cancel "xmcdu::init $modName"

  # Is this module known already?
  if {[info exists MODULES($modName)]} {
    # Module known. Is a registration request pending?
    if {![lindex $MODULES($modName) 1]} {
      # Not any more. Good!
      return
    }
    # Fetch the stored event handler.
    set event [lindex $MODULES($modName) 0]
  } else {
    # Module not known. Does it have a callback?
    if {$event==""} {die "Module $modName needs a callback."}
    # Store its passthrough request.
    if {$passthrough=="passthrough"} {
      set PASSTHROUGH($modName) 1
    } else {
      set PASSTHROUGH($modName) 0
    }
  }; # module known

  # Either a new module, or a repeated registration attempt is required.
  cmd "modadd \{$modName\} [clock seconds]"; # random number for broker kick
  if {$PASSTHROUGH($modName)} {
    cmd "passthrough $modName"
  }

  # Come back at least once for a check.
  set MODULES($modName) [list $event 1]
  after 5000 "xmcdu::init $modName"
  # The pending status will be reset by the MCDU event handler.
}; # init


proc xmcdu::done {{modName {}}} {
  # De-registers from the MCDU.
  # <in> modName = The module name to be de-registered from the MCDU. If
  #                left empty, de-registers all known modules.
  variable MODULES
  variable PASSTHROUGH

  if {$modName=={}} {
    # Disconnect from all known modules.
    foreach module [array names MODULES] {
      cmd "modremove $module"
      after cancel "xmcdu::init $module"
      set pending [lindex $MODULES($module) 1]
      unset MODULES($module)
      unset PASSTHROUGH($module)
      if {!$pending} {
        gui::log xmcdu "De-registered $module subsystem with MCDU"
      }
    }
  } else {
    # Disconnect from the external MCDU.
    cmd "modremove $modName"
    after cancel "xmcdu::init $modName"
    # It is possible that this module does not exist.
    if {![catch {lindex $MODULES($modName) 1} pending]} {
      unset MODULES($modName)
      unset PASSTHROUGH($modName)
      if {!$pending} {
        gui::log xmcdu "De-registered $modName subsystem with MCDU"
      }
    } else {
    }
  }
}; # done


proc xmcdu::available {} {
  # Check whether the external MCDU is on the bus.
  # It depends on the connection method (direct versus Broker) what the
  # exact definition of "available" is.
  # For Broker connections, the MCDU is available if at least one of the
  # requested modules has a "not pending" state.
  # For direct connections, the MCDU is available if the channel is open.
  # <out> 1 if the MCDU is available, otherwise 0.
  variable MODULES
  variable SERVER
  variable MCDU

  if {$SERVER!=""} {
    # Direct connection.
    if {$MCDU==""} {
      return 0
    } else {
      return 1
    }
  } else {
    # Broker connection.
    foreach {module} [array names MODULES] {
      if {[lindex $MODULES($module) 1]==0} {
        return 1
      }
    }
    return 0
  }
}; # available


proc xmcdu::cmd {cmd} {
  # Send a command to the external MCDU.
  # <in> cmd = The command string.
  variable SERVER
  variable MCDU

  if {$SERVER!=""} {
    # Directly talk to the MCDU. This needs local error handling.
    if {[catch {puts $MCDU $cmd}]} {
      # Something went wrong. Don't fumble: report and drop the line, it
      # will be picked up soon enough by checkConnection.
      catch {close $MCDU}
      set MCDU ""
    }
  } else {
    # Talk via the Broker. This includes full error handling.
    # To avoid an error when the application should have been connected to a
    # dedicated server, which happens to be unavailable, wrap in a catch.
    catch {broker::send "put mcdu.cmd \{$cmd\}"}
  }
}; # cmd


proc xmcdu::event {data} {
  # Handle incoming external MCDU events. Typically this procedure is
  # registered with the Broker library to get mcdu.cmd data strings.
  # May be called at any time, also when the xmcdu module is not yet
  # initialised. Events are passed to a module if that is known to be
  # registered, by using the registered event callback handler in the main
  # program:  {event cmd prm1 prm2}.
  variable MODULES

  set cmd  [lindex $data 0]
  set mode [lindex $data 1]
  set prm1 [lindex $data 2]
  set prm2 [lindex $data 3]

  ##### Events that need to be monitored at all times ###################

  # Reset?
  if {$cmd=="key" && $mode=="reset"} {
    # Schedule re-registration attempt for all modules in 3 seconds.
    gui::log xmcdu "MCDU reset detected"
    foreach modName [array names MODULES] {
      set event [lindex $MODULES($modName) 0]
      set MODULES($modName) [list $event 1]
      after 3000 "xmcdu::init $modName"
    }
    return
  }

  # Do we know this module?
  if {[info exists MODULES($mode)]} {
    # Yes we know it. Succesful registration?
    if {$cmd=="modadded"} {
      # Yes, was this module currently pending a registration?
      if {[lindex $MODULES($mode) 1]} {
	# Yes, stop registration attempts.
	set event [lindex $MODULES($mode) 0] 
	set MODULES($mode) [list $event 0]
	gui::log xmcdu "Registered $mode subsystem with MCDU"
      }
      # Always stop processing here. Ignore spurious registration
      # acknowledgments.
      return
    }
    # Forward event minus the mode to the callback event handler.
    set event [lindex $MODULES($mode) 0]
    $event $cmd $prm1 $prm2
  }; # Known module.

  # Totally ignore events for other modules.
  
}; # event


proc xmcdu::connectDirect {{server "localhost:1863"}} {
  # Establish (or check) an existing direct connection to a MCDU.
  # Flow:
  # The application (top level) calls xmcdu::connectDirect to set up a link
  # with the MCDU. The application decides whether to connect direct or via
  # the Broker. xmcdu does not care, as long as either one is available.
  # If the connection fails, xmcdu will keep trying every 30 seconds.
  # If the connection succeeds, everything is set up so that xmcdu can
  # process the MCDU commands.
  # Every 10 seconds, the health of the connection is re-checked and the
  # connection is re-established when needed automatically.
  # <in> server = "host:port" string; defaults are localhost and 1863.
  variable MCDU
  variable SERVER

  if {$MCDU!=""} {
    # Direct connection already available.
    return
  }

  set SERVER $server
  checkConnection
}; # connectDirect


proc xmcdu::disconnectDirect {} {
  # Simply drops the connection. Deregistering needs to be done separately,
  # by a call to xmcdu::done.
  variable MCDU
  variable SERVER

  # Stop the automatic connection retry.
  after cancel xmcdu::checkConnection

  set SERVER ""
  set mcdu $MCDU
  set MCDU ""
  if {![catch {close $mcdu}]} {
    gui::log xmcdu "Disconnected from MCDU"
  }
}; # disconnectDirect


proc xmcdu::eventDirect {ch} {
  # Called by the stream event handler for the Direct Connection.
  # <in> ch = The channel identifier on which something happened.
  variable MCDU

  if {[eof $ch]} {
    # Abort the line.
    set MCDU ""
    catch {close $ch}
    gui::log xmcdu "MCDU connection aborted"
    return
  }

  # Get the input and forward it to the recipient.
  set line [string trim [gets $ch]]
  if {$line!=""} {
    event $line
  }
}; # eventDirect


proc xmcdu::checkConnection {} {
  # Check whether the MCDU connection is sane. If not, try to
  # reopen it after a delay that depends on the severity of the problem.
  variable MCDU
  variable SERVER

  after cancel xmcdu::checkConnection
  # Initially a 10-second delay for checking the connection again.
  set after 10000

  if {$MCDU==""} {
    set hostPort [split $SERVER ":"]
    set host [lindex $hostPort 0]
    set port [lindex $hostPort 1]
    if {$host==""} {set host "localhost"}
    if {$port==""} {set port 1863}
    gui::log xmcdu "Looking for MCDU on $host:$port..."
    update idletasks

    # The following call blocks for a minute if the host is not online. This
    # may be considered a bug.
    if {[catch {socket $host $port} ch]} {
      gui::log xmcdu "Failed to connect to MCDU: $ch"
      # Retry after 5 seconds (MCDU unreachable).
      set after 5000
    } else {
      # Connection made. Initialize the socket and the handler.
      set MCDU $ch
      fconfigure $ch -buffering line -buffersize 2048 -blocking 0
      fileevent $ch readable "xmcdu::eventDirect $ch"
      gui::log xmcdu "Connected to MCDU"
    }
  }

  # Re-check after some delay.
  after $after xmcdu::checkConnection
}; # checkConnection
