# Engravity 747-400 MCDU library. Takes care of detection, keystroke
# interception and injection in the proper place in the software MCDU,
# copycat display of the strings sent to the software MCDU, and
# annunciators.

# Copyright (C) 2007 Jeroen Hoppenbrouwers (hoppie@hoppie.nl).
# This program is free software; more information in the LICENSE file.

namespace eval engr {
  variable PORT ""          ; # Engravity MCDU hardware handle, if any.
  variable DIRTY            ; # Dirty lines: (line) = 0/1, 1=dirty.
  array set DIRTY {}
  variable KEYS             ; # Translation table (engr)=MCDU.
  # Note that the "CDU MENU" keystroke is generated when leaving the native
  # Engravity Setup Menu of the hardware unit. It is ignored as it is not in
  # this table.
  array set KEYS {
    LS1     lsk1l
    LS2     lsk2l
    LS3     lsk3l
    LS4     lsk4l
    LS5     lsk5l
    LS6     lsk6l
    RS1     lsk1r
    RS2     lsk2r
    RS3     lsk3r
    RS4     lsk4r
    RS5     lsk5r
    RS6     lsk6r
    COMM11  initref
    COMM12  rte
    COMM13  deparr
    COMM14  atc
    COMM15  vnav
    COMM21  fix
    COMM22  legs
    COMM23  hold
    COMM24  fmccomm
    COMM25  prog
    COMM26  exec
    COMM31  menu
    COMM32  navrad
    COMM41  prevpage
    COMM42  nextpage
    1       1
    2       2
    3       3
    4       4
    5       5
    6       6
    7       7
    8       8
    9       9
    0       0
    .       .
    +/-     +/-
    A       a
    B       b
    C       c
    D       d
    E       e
    F       f
    G       g
    H       h
    I       i
    J       j
    K       k
    L       l
    M       m
    N       n
    O       o
    P       p
    Q       q
    R       r
    S       s
    T       t
    U       u
    V       v
    W       w
    X       x
    Y       y
    Z       z
    SP      sp
    DEL     del
    /       /
    CLR     clr    
  }
}; # engr


proc engr::init {} {
  # Try to open the Engravity com port.
  variable PORT

  set port [prefs::get engravity.port]
  if {$port=="off" || $port==""} {
    # Don't open.
    set PORT ""
    return
  }

  # Port entered: try to open it.
  if {[catch {open $port "r+"} mcdu]} {
    # Attempt to open port failed. Message and bail out.
    mcdu::message add "engravity init failed"
    set PORT ""
    return
  }

  # Port opened. Register handler according to default Engravity specs.
  fconfigure $mcdu -mode 38400,n,8,1 -buffering line
  set PORT $mcdu
  fileevent $mcdu readable "engr::incoming"

  # Clear screen and flash message.
  clear
  send "L3T" "F4,S,0,CONNECTED TO MCDU PROGRAM"
  after 1000 {engr::clear}
  return
}; # init


proc engr::done {} {
  # Shut down the Engravity MCDU. It looks like the unit does not shut down
  # completely (i.e., backlighting stays on) when the COM port is closed.
  variable PORT
  if {$PORT!=""} {
    after 500
    clear
    ann dspy 0
    ann fail 0
    ann msg 0
    ann ofst 0
    ann exec 0
    after 500
    close $PORT
    set PORT ""
  }
}; # done


proc engr::ann {name state} {
  # Control the Engravity annunciators.
  # <in> name = One of dspy, fail, msg, ofst, exec. Others are ignored.
  # <in> state = 1 for on, 0 for off.
  switch -- $name {
    dspy {send "SID" "DSPY,$state"}
    fail {send "SID" "FAIL,$state"}
    msg  {send "SID" "MSG,$state"}
    ofst {send "SID" "OFST,$state"}
    exec {send "SID" "EXEC,$state"}
  }
}; # ann


proc engr::clear {} {
  # Clear the complete display and cleans out the dirty list.
  variable DIRTY
  send "CLS"
  for {set row 0} {$row<=13} {incr row} {
    set DIRTY($row) 0
  }
}; # clear


proc engr::display {{row ""}} {
  # The routine needs to prevent the Engravity unit from being
  # swamped with line updates.
  # When row is given, only the "dirty" list is updated. When nothing comes
  # in for 50 ms, the list is run over from top to bottom, sending only
  # dirty lines. Note that the clear command has a shortcut implementation.
  variable DIRTY
  if {$row==""} {
    # Clean up the todo list.
    foreach row [lsort -integer [array names DIRTY]] {
      if {$DIRTY($row)} {
        set DIRTY($row) 0
        show $row
        update
      }
    }
  } else {
    if {$row>13} return
    after cancel engr::display
    set DIRTY($row) 1
    after 50 engr::display
    return
  }
}; # display


proc engr::show {row} {
  # Actual formatter.
  if {$row==0} {
    # Title display. TODO: font stuff (mostly page number).
    set line [mcdu::readRow $row]
    set engrLine ""
    foreach {c f} [split $line ""] {
      append engrLine $c
    }
    send "TIT" "F4,S,0,[xlat $engrLine]"
  } elseif {$row==13} {
    # Scratch pad. Never does small font.
    set line [mcdu::readRow $row]
    set engrLine " "
    foreach {c f} [split $line ""] {
      append engrLine $c
    }
    send "SPD" "F4,S,0,[xlat $engrLine]"
  } else {
    # Normal row.
    set line [mcdu::readRow $row]
    set engrLine ""
    foreach {c f} [split $line ""] {
      append engrLine $c
    }
    set engrLine [xlat $engrLine]
    # Hack: odd rows have small font, even rows large font.
    # This can be improved, but not easily. Better to wait for the word from
    # Engravity about better fonts inside their unit (fixed grid!).
    if {[expr $row%2]} {
      set font "F3"
    } else {
      set font "F4"
    }
    # The Engravity unit uses a different row naming style.
    switch $row {
       1 {set engrRow 1S}
       2 {set engrRow 1T}
       3 {set engrRow 2S}
       4 {set engrRow 2T}
       5 {set engrRow 3S}
       6 {set engrRow 3T}
       7 {set engrRow 4S}
       8 {set engrRow 4T}
       9 {set engrRow 5S}
      10 {set engrRow 5T}
      11 {set engrRow 6S}
      12 {set engrRow 6T}
    }
    if {$font=="F3"} {
      # Small font, needs extra processing.
      # Hack: find the first double space from the left and divide the line.
      set p [string first "  " $engrLine]
      if {$p!=-1} {
        # Split the line in a from-the-left and a from-the-right part.
        set leftPart  [string trim [string range $engrLine 0 [expr $p-1]]]
        set rightPart [string trim [string range $engrLine [expr $p+1] end]]
        send "R$engrRow" "$font,S,0,$rightPart  "
        set engrLine "  $leftPart"
      }
    }
    send "L$engrRow" "$font,S,0, $engrLine"
  }
}; # show


proc engr::xlat {string} {
  # Translate a string of Aerowinx FMC font characters into Engravity chars.
  regsub -all {g} $string [format %c 128] string
  regsub -all {!} $string [format %c 127] string
  regsub -all {\*} $string "a" string
  regsub -all {#} $string "p" string
  return $string
}; # xlat


proc checkSum {string} {
  # Compute the NMEA checksum of the string.
  set sum 0
  foreach ch [split $string ""] {
    scan $ch {%c} ch
    set sum [expr $sum^$ch]
  }
  return [format {%02X} $sum]
}; # checkSum


proc engr::incoming {} {
  # Process any incoming line on the Engravity port.
  variable PORT
  variable KEYS

  if {[eof $PORT]} {
    # ?? Should not really happen.
    mcdu::message add "engravity comm lost"
    done
    return
  }
  set line [gets $PORT]

  # Parse the line for message and check sum.
  if {![regexp {^\$(.*)\*(.*)$} $line dummy \
        message checksum]} {
    # Oops?
    return
  }

  # Test for the correct checksum (never mind USB reliability).
  if {$checksum != [checkSum $message]} {
    # Oops?
    return
  }

  # Further parse the message.
  if {![regexp {^PCDU(.*),(.*)$} $message dummy msgID data]} {
    # Oops?
    return
  }

  # MCDU Alive messages are used as watchdog reset. They should come every
  # five seconds. After ten seconds of silence, the watchdog bites.
  if {$msgID=="ALIVE"} {
    after cancel engr::watchdog
    after 10000 engr::watchdog
    return
  }

  # Process key presses. Be careful to ignore any unknown one, in particular
  # the "CDU MENU" fake keypress generated when leaving the Engravity
  # firmware setup menu. We have no reasonable means to flag these events to
  # the user, so we ignore them.
  if {$msgID=="KEY"} {
    if {[info exists KEYS($data)]} {
      mcdu::keyPressed $KEYS($data)
      return
    }
  }

  # What remains is ignored.
}; # incoming


proc engr::send {msgID {data ""}} {
  # Send a message to the MCDU.
  # Format: $PCDU<msgID>,<data1>,<data2>...*CS
  variable PORT

  if {$PORT==""} return

  # Create the chechsummable part of the message string.
  if {$data==""} {
    set toCheck "PCDU$msgID"
  } else {
    set toCheck "PCDU$msgID,$data"
  }
  # Create the full string.
  set toSend "\$$toCheck*[checkSum $toCheck]"
  if {[catch {puts $PORT $toSend}]} {
    mcdu::message add "engravity comm lost"
    done
  }
}; # send


proc engr::watchdog {} {
  # If this triggers, the Engravity unit has been silent for too long.
  mcdu::message add "engravity comm lost"
  done
}; # watchdog


##### INIT CODE ############################################################

# Try to get the unit immediately, so that the MCDU self test etc. is
# displayed on the Engravity unit as well.
engr::init

