# FMC library for MCDU. Links up the PS1 CDU to the external MCDU.
# $Id: fmc.tcl,v 1.10 2007/07/21 18:13:38 hoppie Exp $

# Copyright (C) 2005 Jeroen Hoppenbrouwers <hoppie@hoppie.nl>
# This program is free software; more information in the LICENSE file.


namespace eval fmc {
  variable FMC_MODE_PRESENT 0;   # 1 if PS1 CDU updates are being received.
  variable KEYMAP;               # [mnemonic]=PS1 pushsw poke code.
  variable KEYCOUNTER 0;         # running counter 0..9 for broker fooling.

  # Private (static).
  variable _DSPY_ANN_LIGHTED 0;  # DSPY annunciator lighted by internal logic.

}; # namespace fmc


proc fmc::done {} {
  # No specific cleanup action to do, yet.
}; # done


proc fmc::init {} {
  # Set up some basic things for the whole module.
  variable KEYMAP

  # Fill in the key map.
  set i 32769; # The base for CDU key codes in PS1.
  foreach code {1 2 3 4 5 6 7 8 9 0 null +/- . a b c d e f g h i j k l m
      n o p q r s t u v w x y z sp del / clr initref rte deparr plan vnav
      fix legs hold prog menu navrad prevpage nextpage exec lsk1l lsk2l lsk3l
      lsk4l lsk5l lsk6l lsk1r lsk2r lsk3r lsk4r lsk5r lsk6r} {
    set KEYMAP($code) $i
    incr i
  }

}; # init


proc fmc::reset {} {
  # Initialise the communication with the PS1 CDU.
  # Send the LEGS and PROG keystrokes to put the PS1 FMC in a known state.
  # Later this may be replaced by a specific 'refresh' function.
  keystroke legs
  after 100 {fmc::keystroke prog}
}; # reset


proc fmc::crtEvent {event} {
  # Receives all CRT events on the PS1 CDU, via the ps1.cdu.crt key.
  # <in> event = The event as specified by the VMD CRT output. Either
  #              "end" or "line_nr data line_nr data", see fmc::display.
  variable FMC_MODE_PRESENT;

  if {$event=="end"} {
    # Apparently the CDU reader quit. Withdraw the FMC mode from the MCDU.
    # Play this via the Broker, to let other MCDUs know about it, too.
    after 200 {ps1::toBroker "put mcdu.cmd {modremove fmc}"}
    set FMC_MODE_PRESENT 0
  } else {
    # Is the FMC mode already present? Check locally for speed reasons.
    if {$FMC_MODE_PRESENT==0} {
      # Not present yet, add it to the MCDU menu via the Broker.
      after 2000 {
        ps1::toBroker "put mcdu.cmd {modadd fmc}"
        ps1::toBroker "put mcdu.cmd {modkey fmc {initref rte deparr vnav fix legs hold prog exec navrad}}"
      }
    }
    # Just to stay synchronised.
    set FMC_MODE_PRESENT 1
    # If in FMC mode, pass event on to the displayer.
    if {[mcdu::mode]=="fmc"} {
      fmc::display $event
    }
  }
}; # crtEvent


proc fmc::annEvent {key value} {
  # Receives all annunciator events on the PS1 CDU.
  # <in> key = The VMread mnemonic of the key.
  # <in> value = The value of the key.

  # If not in FMC mode, don't do anything.
  if {[mcdu::mode]!="fmc"} {
    return
  }

  switch -- $key {
    VM.MCDUEXELight {
      if {$value==0} {
	mcdu::ann exec 0
      } else {
	mcdu::ann exec 1
      }
    }
    VM.MCDUMsgLight {
      if {$value==0} {
	mcdu::ann msg 0
      } else {
	mcdu::ann msg 1
      }
    }
    VM.MCDUOffsetLight {
      if {$value==0} {
	mcdu::ann ofst 0
      } else {
	mcdu::ann ofst 1
      }
    }
  }; # switch

}; # annEvent


proc fmc::display {data} {
  # Receives all CRT changes of the PS1 CDU. The data comes in per line,
  # possibly more than one line per packet. This procedure matches incoming
  # lines to a list of known anomalies, and tries to fix them.
  # <in> data = {row_nr hex_line_data row_nr hex_line_data ...}
  variable _DSPY_ANN_LIGHTED

  foreach {row text} $data {
    # Fix match? Can be done more elaborately by also matching the page
    # title, but at this moment this does not seem to be required.
    regsub {202020202020212121272121212020212121212721212020} $text \
           {2020202020202121212721212E2120212121212721212E21} text
    regsub {ADADADADADADADADADADADADADADADADAD2D2D2D2D2D2D2D} $text \
           {ADADADADADADADADADADADADADADADADADADADADADADADAD} text
    regsub {21212020$} $text {21212121} text

    # Decode hex.
    set text [binary format H* $text]

    # Display the whole row, column (character) for column.
    # Keep a row string variable for subsequent pattern matching.
    set rowString ""
    for {set col 0} {$col<24} {incr col} {
      set char [string index $text $col]
      if {$char>"z"} {
        # Is a small font character. Strip 7th bit (MSB).
        scan $char "%c" code
        incr code -128
        set char [format "%c" $code]
	set font "small"
      } else {
	set font "large"
      }
      mcdu::display $row $col $char $font
      append rowString $char
    }; # for each column

    # A few cases of PS1.3 forgetting to display an annunciator. These are
    # only relevant when displaying a title line (row 0) which is not empty.
    if {$row==0 && $rowString!="                        "} {
      # By default, the DSPY annunciator is OFF.
      set dspyOnNow 0

      # If title matches "ACT RTE" and page is not 1/?, light DSPY.
      if {[regexp {ACT RTE [12]} $rowString]} {
        if {[regexp {[0-9 ][^1]/[0-9][0-9]?} $rowString]} {
          set dspyOnNow 1
        }
      }
      # If title matches "MOD RTE", light DSPY.
      if {[regexp {MOD RTE [12]} $rowString]} {
        set dspyOnNow 1
      }

      # Actual annunciator control takes place here. If it was on the last
      # time we checked, and it should not be on now, distinghuish it.
      if {$_DSPY_ANN_LIGHTED && !$dspyOnNow} {
        mcdu::ann dspy 0
        set _DSPY_ANN_LIGHTED 0
      }
      # If it was not on the last time we checked, and it should be on now,
      # illuminate it.
      if {!$_DSPY_ANN_LIGHTED && $dspyOnNow} {
        mcdu::ann dspy 1
        set _DSPY_ANN_LIGHTED 1
      }
      # The other two cases (on-on, off-off) need no action.
      
    }; # if row==0

  }; # for each row

}; # display


proc fmc::keystroke {key} {
  # Sends the given keystroke to the PS1 CDU, using a low-level interface.
  # The KEYCOUNTER is only used to enable getting repeated keystrokes through
  # the Broker, by producing a different value each time.
  # <in> key = Mnemonical keystroke, standard MCDU mnemonics (see mcdu.tcl).
  global ETCDIR
  variable KEYMAP
  variable KEYCOUNTER

  # Translate mnemonic to ordinal "poke" value, if a meaningful one exists.
  if {![info exists KEYMAP($key)]} return
  set keystroke $KEYMAP($key)

  # Send to Broker.
  ps1::toBroker "put IPC.pushsw {$keystroke 1 $KEYCOUNTER}"
  set KEYCOUNTER [expr ($KEYCOUNTER+1)%10]
}; # keystroke


##### Module startup code #################################################

fmc::init
