# Connector module for PS/1.3.
# $Id: ps13.tcl,v 1.7 2006/04/02 15:12:23 hoppie Exp $

# Copyright (C) 2005 Jeroen Hoppenbrouwers <hoppie@hoppie.nl>
# This program is free software; more information in the LICENSE file.


# This module is used to send and receive formatted ASCII messages to and from
# the MCDU, usually via the PS1.3 Broker, but as an option also directly via a
# built-in socket server. For historical and compatibility reasons, there is
# much reference to PS1, while there actually is little PS1-dependency in this
# module at all.
# While more than one connection to the local server is supported, it is not
# possible to link more than one MCDU in the local server configuration.

##### constants ###########################################################

namespace eval ps1 {
  variable PORT [expr 0x747] ; # Broker port, init is 1863, can be changed
  variable BROKER 0          ; # Broker channel handle, 0 = invalid
  variable CHANNELS          ; # Array with all open channels in server mode
  array set CHANNELS {}
}

##### procs ###############################################################


proc ps1::done {} {
  # Neatly disconnect from the Broker.
  toBroker "exit"
}


proc ps1::log {msg} {
  # Show a message on the MCDU.
  # Suppress a successful Broker logon (would be a nuisance).
  if {[string range $msg 0 22]=="Connected to 747 Broker"} return
  mcdu::message add $msg
}; # log


proc ps1::toBroker {msg} {
  # Send the message to the Broker. Channel may be down, no problem.
  # In case the Broker host is set to "server", only send the actual MCDU
  # event, not the "put mcdu.cmd" Broker key. Note that the local server
  # may serve more than one actual channel.
  # <in> msg = The message (line) to send.
  # <out> 0 on success, 1 on failure to send.
  variable BROKER
  variable CHANNELS

  if {[string range [prefs::get broker.host] 0 5]=="server"} {
    # Local server. No Broker commands required, strip them.
    set msg [lindex $msg 2]
    # Send the event to each open channel.
    foreach ch [array names CHANNELS] {
      catch {puts $ch $msg}
    }
    return 0
  }

  # Broker connection.
  if {[catch {puts $BROKER $msg}]} {
    return 1
  } else {
    return 0
  }
}; # toBroker


proc ps1::checkBrokerConnection {} {
  # Check whether the Broker connection is sane. If not, try to
  # reopen it after a delay that depends on the severity of the problem.
  variable PORT
  variable BROKER
  variable CHANNELS

  # Stop off a possible parallel try loop.
  after cancel ps1::checkBrokerConnection

  # Initially a 1-second delay for checking the connection again.
  set after 1000

  # If no connection at this moment, try something.
  if {$BROKER==0} {
    # Get host IP. It may contain a /port element.
    set host [prefs::get broker.host]
    set pos [string first "/" $host]
    if {$pos!=-1} {
      incr pos
      set PORT [string range $host $pos end]
      incr pos -2
      set host [string range $host 0 $pos]
    } else {
      set PORT [expr 0x747]
    }
    if {$host==""} {set host "off"}
    if {$host=="off"} {
      # Do not retry until activated again.
      mcdu::ann fail 0
      return
    }
    if {$host=="server"} {
      # Start the local socket server on the same port as the Broker would
      # be.
      startServer
      return
    }

    update idletasks

    # Try to open a socket.
    if {[catch {socket $host $PORT} ch]} {
      log "Broker unavailable"
      mcdu::ann fail 1
      # Retry after 15 seconds (Broker unreachable).
      set after 15000
    } else {
      # Connection made. Initialize the socket and the handler.
      set BROKER $ch
      fconfigure $ch -buffering line -buffersize 2048 -blocking 0
      fileevent $ch readable "ps1::readEvent $ch"

      # Broker now up. Register MCDU.
      toBroker "register {MCDU} {Online}"
      # Empty the current mcdu.cmd/event keys, just in case.
      toBroker "put mcdu.cmd { } mcdu.event { }"
      # Register an interest in the mcdu.cmd key.
      toBroker "notify mcdu.cmd"
      # Register an interest in the relevant PS1 CDU keys.
      toBroker "notify ps1.cdu.crt VM.MCDUEXELight VM.MCDUMsgLight \
                       VM.MCDUOffsetLight VM.StandByPower"
      # Remove a possible message/FAIL annunciator.
      mcdu::message remove "Broker unavailable"
      mcdu::ann fail 0
    }
  }

  # Re-check after some delay.
  after $after ps1::checkBrokerConnection
}; # checkBrokerConnection


proc ps1::readEvent {ch} {
  # Incoming message on the socket. In case of local server connection,
  # expect direct MCDU commands instead of Broker mcdu.cmd keys.
  variable BROKER
  variable CHANNELS

  if {[eof $ch]} {
    # Close the connection.
    catch {close $ch}
    if {[string range [prefs::get broker.host] 0 5]=="server"} {
      # Local server.
      unset CHANNELS($ch)
    } else {
      # Broker connection.
      set BROKER 0
    }
    return
  }

  if {[catch {set line [gets $ch]}]} return
  if {$line==""} return

  # All set up, process the incoming command(s) and data.
  if {[string range [prefs::get broker.host] 0 5]=="server"} {
    # Direct connection, no Broker commands, direct MCDU command.
    mcdu::cmd $line
  } else {
    # Broker connection, check for mcdu.cmd and other keys.
    switch -- [lindex $line 0] {
      data {
	foreach {key value} [lrange $line 1 end] {
	  switch -glob -- $key {
	    mcdu.cmd {
	      if {$value!=""} {
		mcdu::cmd $value
	      }
	    }
	    ps1.cdu.crt {
	      fmc::crtEvent $value
	    }
	    VM.MCDU*Light {
	      fmc::annEvent $key $value
	    }
            VM.StandByPower {
              mcdu::powerEvent $value
            }
	  }; # switch
	}; # foreach key/value pair
      }
      default {
	log $line
      }
    }; # switch
  }; # if local server
}; # readEvent


proc ps1::startServer {} {
  # Routine to start a local socket server, in order to bypass the Broker
  # and, essentially, offer a direct connection for interested programs to
  # control the MCDU.
  variable PORT
  # Start a socket server.
  if {[catch {socket -server ps1::server $PORT} ss]} {
    log "$ss"
  }
}; # startServer


proc ps1::server {ch host port} {
  # Called when a new connection is initiated by a client. Not used for
  # Broker communications at all.
  variable CHANNELS
  fconfigure $ch -buffering line -buffersize 65535 -blocking 0
  fileevent $ch readable "ps1::readEvent $ch"
  set CHANNELS($ch) ""
}; # server


##### init ################################################################

after 1000 ps1::checkBrokerConnection
